﻿using System.Windows.Forms;

namespace HeuristicLab.BenchmarkSuite.Views {

  using HeuristicLab.BenchmarkSuite.Problems;
  using HeuristicLab.Core.Views;
  using HeuristicLab.MainForm;

  [View("Push Expression Selection Editor")]
  [Content(typeof(Data), true)]
  public partial class DataEditorView : NamedItemView {
    private const string ValueSeparator = ", ";

    public DataEditorView() {
      InitializeComponent();

      dataGridView.CellEndEdit += DataGridViewCellEndEdit;
    }

    private void DataGridViewCellEndEdit(object sender, DataGridViewCellEventArgs e) {
      if (Content == null)
        return;

      var row = dataGridView.Rows[e.RowIndex];

      var inputArgs = new string[Content.InputArgumentTypes.Length];
      for (var i = 0; i < Content.InputArgumentTypes.Length; i++) {
        inputArgs[i] = row.Cells[i].Value.ToString();
      }

      var outputArgs = new string[Content.OutputArgumentTypes.Length];
      for (var i = 0; i < Content.OutputArgumentTypes.Length; i++) {
        outputArgs[i] = row.Cells[Content.InputArgumentTypes.Length + i].Value.ToString();
      }

      var example = Content.Examples[e.RowIndex];

      try {
        Content.Examples[e.RowIndex] = ParseExample(inputArgs, Content.InputArgumentTypes, outputArgs, Content.OutputArgumentTypes);
      }
      catch {
        MessageBox.Show("Unable to parse example!");

        // reset entry
        row.Cells[e.ColumnIndex].Value = e.ColumnIndex < Content.InputArgumentTypes.Length
          ? example.InputArgs[e.ColumnIndex]
          : example.OutputArgs[e.ColumnIndex - Content.InputArgumentTypes.Length];
      }
    }

    private Example ParseExample(string[] inputArgs, ExampleArgumentType[] inputArgTypes, string[] outputArgs, ExampleArgumentType[] outputArgTypes) {
      var example = new Example();

      example.InputArgs = new string[inputArgs.Length];
      for (var i = 0; i < inputArgs.Length; i++) {
        var arg = inputArgs[i];
        var type = inputArgTypes[i];

        switch (type) {
          case ExampleArgumentType.Bool:
            example.InputBoolean = new[] { ExampleArgumentConverter.ConvertBoolean(arg) };
            example.InputArgs[i] = arg;
            break;
          case ExampleArgumentType.Integer:
            example.InputInt = new[] { ExampleArgumentConverter.ConvertInteger(arg) };
            example.InputArgs[i] = arg;
            break;
          case ExampleArgumentType.IntegerCollection:
            example.InputInt = ExampleArgumentConverter.ConvertIntegers(arg);
            example.InputArgs[i] = string.Join(ValueSeparator, example.InputInt);
            break;
          case ExampleArgumentType.Float:
            example.InputFloat = new[] { ExampleArgumentConverter.ConvertDouble(arg) };
            example.InputArgs[i] = arg;
            break;
          case ExampleArgumentType.FloatCollection:
            example.InputFloat = ExampleArgumentConverter.ConvertDoubles(arg);
            example.InputArgs[i] = string.Join(ValueSeparator, example.InputFloat);
            break;
        }
      }

      example.OutputArgs = new string[outputArgs.Length];
      for (var i = 0; i < outputArgs.Length; i++) {
        var arg = outputArgs[i];
        var type = outputArgTypes[i];

        switch (type) {
          case ExampleArgumentType.Bool:
            example.OutputBoolean = new[] { ExampleArgumentConverter.ConvertBoolean(arg) };
            example.OutputArgs[i] = arg;
            break;
          case ExampleArgumentType.Integer:
            example.OutputInt = new[] { ExampleArgumentConverter.ConvertInteger(arg) };
            example.OutputArgs[i] = arg;
            break;
          case ExampleArgumentType.IntegerCollection:
            example.OutputInt = ExampleArgumentConverter.ConvertIntegers(arg);
            example.OutputArgs[i] = string.Join(ValueSeparator, example.OutputInt);
            break;
          case ExampleArgumentType.Float:
            example.OutputFloat = new[] { ExampleArgumentConverter.ConvertDouble(arg) };
            example.OutputArgs[i] = arg;
            break;
          case ExampleArgumentType.FloatCollection:
            example.OutputFloat = ExampleArgumentConverter.ConvertDoubles(arg);
            example.OutputArgs[i] = string.Join(ValueSeparator, example.OutputFloat);
            break;
        }
      }

      return example;
    }

    public new Data Content
    {
      get { return (Data)base.Content; }
      set
      {
        base.Content = value;
      }
    }

    protected override void OnReadOnlyChanged() {
      base.OnReadOnlyChanged();

      dataGridView.Enabled = !ReadOnly;
    }

    protected override void OnContentChanged() {
      dataGridView.Columns.Clear();
      dataGridView.Rows.Clear();

      if (Content == null) {
        return;
      }

      this.nameTextBox.Text = Content.Name;

      var cellTemplate = new DataGridViewTextBoxCell();

      for (var i = 0; i < Content.InputArgumentTypes.Length; i++) {
        var headerTypeName = ViewHelper.GetHeaderTypeName(Content.InputArgumentTypes[i]);
        var column = new DataGridViewColumn {
          HeaderText = string.Format("Input {0} : {1}", i + 1, headerTypeName),
          AutoSizeMode = DataGridViewAutoSizeColumnMode.Fill,
          CellTemplate = cellTemplate
        };

        dataGridView.Columns.Add(column);
      }

      for (var i = 0; i < Content.OutputArgumentTypes.Length; i++) {
        var headerTypeName = ViewHelper.GetHeaderTypeName(Content.OutputArgumentTypes[i]);
        var column = new DataGridViewColumn {
          HeaderText = string.Format("Output {0} : {1}", i + 1, headerTypeName),
          AutoSizeMode = DataGridViewAutoSizeColumnMode.Fill,
          CellTemplate = cellTemplate
        };

        dataGridView.Columns.Add(column);
      }

      foreach (var example in Content.Examples) {
        var row = new DataGridViewRow();
        row.HeaderCell.Value = row.Index + 1;

        row.CreateCells(dataGridView);

        for (var i = 0; i < Content.InputArgumentTypes.Length; i++) {
          row.Cells[i].Value = ViewHelper.StringifyInput(Content.InputArgumentTypes[i], example, ValueSeparator);
        }

        for (var i = 0; i < Content.OutputArgumentTypes.Length; i++) {
          row.Cells[Content.InputArgumentTypes.Length + i].Value = ViewHelper.StringifyOutput(Content.OutputArgumentTypes[i], example, ValueSeparator);
        }

        dataGridView.Rows.Add(row);
      }
    }
  }
}
