﻿using System;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Collections.Generic;
  using System.Linq;
  using BenchmarkSuite.Problems;
  using Common;
  using Configuration;
  using Core;
  using Data.Pool;
  using Encodings.IntegerVectorEncoding;
  using Expressions;
  using HeuristicLab.Data;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Erc;

  using Instances;
  using Interpreter;
  using Optimization;
  using Parameters;
  using Persistence.Default.CompositeSerializers.Storable;
  using Random;
  using Stack;

  [StorableClass]
  [Creatable(CreatableAttribute.Categories.GeneticProgrammingProblems, Priority = 180)]
  [Item("Push Problem", "")]
  public class PushProblem : SingleObjectiveBasicProblem<IntegerVectorEncoding>, IProblemInstanceConsumer<Data> {
    [Storable]
    private readonly PushConfiguration config;
    private PushInterpreterPool pool;
    private readonly ObjectPool<IRandom> randomPool = new ObjectPool<IRandom>(() => new MersenneTwister());

    public const string CaseQualitiesScopeParameterName = "CaseQualities";

    public PushProblem() {
      config = new PushConfiguration();

      InitData();
      InitEvents();
      InitParameters();
    }

    [StorableConstructor]
    public PushProblem(bool deserializing)
      : base(deserializing) {
    }

    public PushProblem(PushProblem original, Cloner cloner)
      : base(original, cloner) {
      config = cloner.Clone(original.config);

      InitData();
      InitEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitData();
      InitEvents();
    }

    private void InitData() {
      pool = new PushInterpreterPool(Environment.ProcessorCount * 2, 4096, 1024, config);

      //var solutionCreator = new PointsBasedPushProgramCreator(config.ErcOptions);
      //SolutionCreator = solutionCreator;
      //Encoding.SolutionCreator = solutionCreator;
    }

    private void InitEvents() {
      config.EnabledExpressionsChanged += EnabledExpressionsChanged;
    }

    private void EnabledExpressionsChanged(object sender, EnabledExpressionsChangedEventArgs e) {
      this.Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      this.Encoding.BoundsParameter.Value[0, 1] = config.EnabledExpressions.Count;
    }

    #region Parameters

    private const string DataBoundsParameterName = "DataBounds";
    private const string DataParameterName = "Data";
    private const string DataParameterDescription = "Program Synthesis";
    private const string InstructionsParameterName = "Instructions";
    private const string InstructionsParameterDescription = "Enables/Disables Instructions";
    private const string EvalPushLimitParameterName = "EvalPushLimit";
    private const string EvalPushLimitParameterDescription = "This is the maximum allowed number of \"executions\" in a single top-level call to the interpreter. The execution of a single Push instruction counts as one execution, as does the processing of a single literal, as does the descent into one layer of parentheses (that is, the processing of the \"(\" counts as one execution).";
    private const string MaxPointsInProgramParameterName = "MaxProgramLength";
    private const string MaxProgramLengthParameterDescription = "This is the maximum size of an item on the CODE stack, expressed as a number of points. A point is an instruction, a literal, or a pair of parentheses.";
    private const string TopLevelPushCodeParameterName = "TopLevelPushCode";
    private const string TopLevelPushCodeParameterDescription = "When TRUE (which is the default), code passed to the top level of the interpreter will be pushed onto the CODE stack prior to execution.";
    private const string TopLevelPopCodeParameterName = "TopLevelPopCode";
    private const string TopLevelPopCodeParameterDescription = "When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.";
    private const string MinRandomIntegerParameterName = "MinRandomInteger";
    private const string MinRandomIntegerParameterDescription = "The minimum INTEGER that will be produced as an ephemeral random INTEGER constant or from a call to INTEGER.RAND.";
    private const string MaxRandomIntegerParameterName = "MaxRandomInteger";
    private const string MaxRandomIntegerParameterDescription = "The maximum INTEGER that will be produced as an ephemeral random INTEGER constant or from a call to INTEGER.RAND.";
    private const string MinRandomFloatParameterName = "MinRandomFloat";
    private const string MinRandomFloatParameterDescription = "The minimum FLOAT that will be produced as an ephemeral random FLOAT constant or from a call to FLOAT.RAND.";
    private const string MaxRandomFloatParameterName = "MaxRandomFloat";
    private const string MaxRandomFloatParameterDescription = "The maximum FLOAT that will be produced as an ephemeral random FLOAT constant or from a call to FLOAT.RAND.";
    private const string NewErcNameProbabilityParameterName = "NewErcNameProbability";
    private const string NewErcNameProbabilityParameterDescription = "The probability that the selection of the ephemeral random NAME constant for inclusion in randomly generated code will produce a new name.";
    private const string MaxPointsInRandomInstructionParameterName = "MaxPointsInRandomInstruction";
    private const string MaxPointsInRandomInstructionParameterDescription = "MaxPointsInRandomInstruction";
    private const string ErcOptionsParameterName = "ERC options";
    private const string MaxStringLengthParameterName = "Max. string length";
    private const string MaxDepthParameterName = "Max. program recursion";

    private void InitParameters() {
      Parameters.Add(new FixedValueParameter<DataBounds>(DataBoundsParameterName));

      Parameters.Add(new ValueParameter<IEnabledExpressionsConfiguration>(
        InstructionsParameterName,
        InstructionsParameterDescription,
        config));

      Parameters.Add(new ValueParameter<Data>(
        DataParameterName,
        DataParameterDescription));

      Parameters.Add(new FixedValueParameter<ErcOptions>(ErcOptionsParameterName, config.ErcOptions));

      Parameters.Add(new FixedValueParameter<IntValue>(
        EvalPushLimitParameterName,
        EvalPushLimitParameterDescription,
        new IntValue(config.EvalPushLimit)));

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxPointsInProgramParameterName,
        MaxProgramLengthParameterDescription,
        new IntValue(config.MaxPointsInProgram)));
      Encoding.LengthParameter = Parameters[MaxPointsInProgramParameterName] as IFixedValueParameter<IntValue>;

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPushCodeParameterName,
        TopLevelPushCodeParameterDescription,
        new BoolValue(config.TopLevelPushCode)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<BoolValue>(
        TopLevelPopCodeParameterName,
        TopLevelPopCodeParameterDescription,
        new BoolValue(config.TopLevelPopCode)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxPointsInRandomInstructionParameterName,
        MaxPointsInRandomInstructionParameterDescription,
        new IntValue(config.MaxPointsInRandomExpression)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxStringLengthParameterName,
        new IntValue(config.MaxStringLength)) { Hidden = true });

      Parameters.Add(new FixedValueParameter<IntValue>(
        MaxDepthParameterName,
        new IntValue(config.MaxDepth)) { Hidden = true });

      Parameters.Add(new LookupParameter<BoolArray>("Cases", "The training cases that have been successfully executed."));
      Parameters.Add(new LookupParameter<DoubleArray>(CaseQualitiesScopeParameterName, "The quality of every single training case for each individual"));

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.Length = config.MaxPointsInProgram;
    }

    public IValueParameter<DataBounds> DataBoundsParameter
    {
      get { return (IValueParameter<DataBounds>)Parameters[DataBoundsParameterName]; }
    }

    public DataBounds DataBounds
    {
      get { return DataBoundsParameter.Value; }
      set { DataBoundsParameter.Value = value; }
    }


    public IValueParameter<IEnabledExpressionsConfiguration> InstructionsParameter
    {
      get { return (IValueParameter<IEnabledExpressionsConfiguration>)Parameters[InstructionsParameterName]; }
    }

    public IEnabledExpressionsConfiguration Instructions
    {
      get { return InstructionsParameter.Value; }
      set { InstructionsParameter.Value = value; }
    }

    public IValueParameter<Data> DataParameter
    {
      get { return (IValueParameter<Data>)Parameters[DataParameterName]; }
    }

    public Data Data
    {
      get { return DataParameter.Value; }
      set { DataParameter.Value = value; }
    }

    public IValueParameter<ErcOptions> ErcOptionsParameter
    {
      get { return (IValueParameter<ErcOptions>)Parameters[ErcOptionsParameterName]; }
    }

    public ErcOptions ErcOptions
    {
      get { return config.ErcOptions; }
      set
      {
        ErcOptionsParameter.Value = value;
        config.ErcOptions = value;
      }
    }

    /// <summary>
    ///     This is the maximum allowed number of "executions" in a single top-level call to the interpreter.
    ///     The execution of a single Push instruction counts as one execution, as does the processing of a single literal,
    ///     as does the descent into one layer of parentheses (that is, the processing of the "(" counts as one execution).
    ///     When this limit is exceeded the interpreter aborts immediately, leaving its stacks in the states they were in prior
    ///     to the abort (so they may still be examined by a calling program). Whether or not this counts as an "abnormal"
    ///     termination
    ///     is up to the calling program.
    /// </summary>
    public IValueParameter<IntValue> EvalPushLimitParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[EvalPushLimitParameterName]; }
    }

    public int EvalPushLimit
    {
      get { return config.EvalPushLimit; }
      set
      {
        EvalPushLimitParameter.Value.Value = value;
        config.EvalPushLimit = value;
      }
    }

    /// <summary>
    /// This is the maximum of depth a push program can have. Expressions, which lead to exceed this limit are interpreted as NOOP.
    /// </summary>
    public IValueParameter<IntValue> MaxDepthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxDepthParameterName]; }
    }

    public int MaxDepth
    {
      get { return config.MaxDepth; }
      set
      {
        MaxDepthParameter.Value.Value = value;
        config.MaxDepth = value;
      }
    }

    /// <summary>
    ///     This is the maximum size of an item on the CODE stack, expressed as a number of points.
    ///     A point is an instruction, a literal, or a pair of parentheses. Any instruction that would cause this limit to be
    ///     exceeded
    ///     should instead act as a NOOP, leaving all stacks in the states that they were in before the execution of the
    ///     instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInProgramParameter
    {
      get { return (IValueParameter<IntValue>)this.Parameters[MaxPointsInProgramParameterName]; }
    }

    public int MaxPointsInProgram
    {
      get { return config.MaxPointsInProgram; }
      set
      {
        MaxPointsInProgramParameter.Value.Value = value;
        config.MaxPointsInProgram = value;
      }
    }

    /// <summary>
    ///     The maximum number of points in an expression produced by the CODE.RAND instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInRandomExpressionParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInRandomInstructionParameterName]; }
    }

    public int MaxPointsInRandomExpression
    {
      get { return config.MaxPointsInRandomExpression; }
      set
      {
        MaxPointsInRandomExpressionParameter.Value.Value = value;
        config.MaxPointsInRandomExpression = value;
      }
    }

    /// <summary>
    ///     When TRUE (which is the default), code passed to the top level of the interpreter
    ///     will be pushed onto the CODE stack prior to execution.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPushCodeParameter
    {
      get { return (IValueParameter<BoolValue>)this.Parameters[TopLevelPushCodeParameterName]; }
    }

    public bool TopLevelPushCode
    {
      get { return config.TopLevelPushCode; }
      set
      {
        TopLevelPushCodeParameter.Value.Value = value;
        config.TopLevelPushCode = value;
      }
    }

    /// <summary>
    ///     When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPopCodeParameter
    {
      get { return (IValueParameter<BoolValue>)this.Parameters[TopLevelPopCodeParameterName]; }
    }

    public bool TopLevelPopCode
    {
      get { return config.TopLevelPopCode; }
      set
      {
        TopLevelPopCodeParameter.Value.Value = value;
        config.TopLevelPopCode = value;
      }
    }

    public IValueParameter<IntValue> MaxStringLengthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxStringLengthParameterName]; }
    }

    public int MaxStringLength
    {
      get { return config.MaxStringLength; }
      set
      {
        MaxStringLengthParameter.Value.Value = value;
        config.MaxStringLength = value;
      }
    }
    #endregion

    public override bool Maximization { get { return false; } }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PushProblem(this, cloner);
    }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      const string bestSolutionResultName = "Best Solution";
      var bestQuality = Maximization ? qualities.Max() : qualities.Min();
      var bestIdx = Array.IndexOf(qualities, bestQuality);
      var bestIndividual = individuals[bestIdx];
      var solution = new PushSolution(bestIndividual.IntegerVector(), bestQuality, Data, random, config, DataBounds.TrainingRange.Start, DataBounds.TrainingRange.End);

      if (!results.ContainsKey(bestSolutionResultName)) {
        results.Add(new Result(bestSolutionResultName, solution));
      } else {
        var currentBestQuality = ((PushSolution)results[bestSolutionResultName].Value).Quality;

        if (Maximization && currentBestQuality < bestQuality ||
           !Maximization && currentBestQuality > bestQuality) {
          results[bestSolutionResultName].Value = solution;
        }
      }
    }

    public override double Evaluate(Individual individual, IRandom random) {
      var program = individual.ToPushProgram(config, randomPool);
      var result = PushEvaluator.Evaluate(
        program,
        pool,
        random,
        Data,
        DataBounds.TrainingRange.Start,
        DataBounds.TrainingRange.End);

      individual.SetScopeValue(CaseQualitiesScopeParameterName, new DoubleArray(result.ExampleQualities));

      return result.TotalQuality;
    }

    public void Load(Data data) {
      Data = data;
      BestKnownQuality = data.BestResult;
      MaxPointsInProgram = data.MaxSize;
      EvalPushLimit = data.EvalLimit;

      config.EnabledExpressions = (IList<string>)ExpressionTable.GetExpressionsByStackTypes((StackTypes)data.EnabledDataTypes);

      // update enabled stack types
      foreach (var stackType in ExpressionTable.StackTypeToNamesTable.Keys) {
        var enable = config.EnabledExpressions.Intersect(ExpressionTable.StackTypeToNamesTable[stackType]).Any();
        config.SetStack(stackType, enable);
      }

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.Length = config.MaxPointsInProgram;

      DataBounds.TrainingRange.Start = 0;
      DataBounds.TrainingRange.End = Data.OriginalTrainingCount;
      DataBounds.TestRange.Start = Data.OriginalTrainingCount;
      DataBounds.TestRange.End = Data.OriginalTrainingCount + Data.OriginalTestCount;
    }
  }
}